/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1985 Gary W. Ng
         1992 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include <math.h>
#include "diodefs.h"
#include "fteconst.h"
#include "iferrmsg.h"
#include "noisedef.h"
#include "util.h"

/*
 * DIOnoise (mode, operation, firstModel, ckt, data, OnDens)
 *     This routine names and evaluates all of the noise sources
 *     associated with diodes.  It starts with the model *firstModel and
 *     traverses all of its instancess.  It then proceeds to any other
 *     models on the linked list.  The total output noise density
 *     generated by all of the diodes is summed with the variable
 *     "OnDens".
 */


/* define the names of the noise sources */

/* Note that we have to keep the order
 * consistent with the index definitions
 * in DIOdefs.h
 */
static char *DIOnNames[DIONSRCS] = {
    ".rs",              /* noise due to rs */
    ".id",              /* noise due to id */
    ".1overf",          /* flicker (1/f) noise */
    ""                  /* total diode noise */
};


int
DIOnoise (mode, operation, genmodel, ckt, dataptr, OnDens)

int mode;
int operation;
GENmodel *genmodel;
CKTcircuit *ckt;
GENERIC *dataptr;
double *OnDens;
{
    Ndata *data = (Ndata*)dataptr;
    DIOmodel *model = (DIOmodel *) genmodel;
    DIOinstance *inst;
    double tempOnoise;
    double tempInoise;
    double noizDens[DIONSRCS];
    double lnNdens[DIONSRCS];
    int error;
    int i;
    char name[N_MXVLNTH];

    if (operation == N_OPEN) {

        /* see if we have to to produce a summary report
         * if so, name all the noise generators
         */

        if (((NOISEAN*)ckt->CKTcurJob)->NStpsSm == 0)
            return (OK);

        if (mode == N_DENS) {

            for ( ; model; model = model->DIOnextModel) {
                for (inst = model->DIOinstances; inst;
                        inst = inst->DIOnextInstance) {

                    for (i = 0; i < DIONSRCS; i++) {
                        (void)sprintf(name,
                            "onoise.%s%s",inst->DIOname,DIOnNames[i]);

                        data->namelist = (IFuid *)
                            trealloc((char *)data->namelist,
                                (data->numPlots + 1)* sizeof(IFuid));

                        if (!data->namelist)
                            return (E_NOMEM);

                        (*(SPfrontEnd->IFnewUid))(ckt,
                            &(data->namelist[data->numPlots++]),
                            (IFuid)NULL,name,UID_OTHER,(GENERIC **)NULL);
                        /* we've added one more plot */
                    }
                }
            }
            return (OK);
        }

        if (mode == INT_NOIZ) {

            for ( ; model; model = model->DIOnextModel) {
                for (inst = model->DIOinstances; inst;
                        inst = inst->DIOnextInstance) {

                    for (i = 0; i < DIONSRCS; i++) {
                        (void)sprintf(name,
                            "onoise_total.%s%s",inst->DIOname,DIOnNames[i]);

                        data->namelist = (IFuid *)
                            trealloc((char *)data->namelist,
                                (data->numPlots + 1)*sizeof(IFuid));

                        if (!data->namelist)
                            return (E_NOMEM);

                        (*(SPfrontEnd->IFnewUid))(ckt,
                            &(data->namelist[data->numPlots++]),
                            (IFuid)NULL,name,UID_OTHER,(GENERIC **)NULL);
                        /* we've added one more plot */

                        (void)sprintf(name,
                            "inoise_total.%s%s",inst->DIOname,DIOnNames[i]);

                        data->namelist = (IFuid *)
                            trealloc((char *)data->namelist,
                                (data->numPlots + 1)*sizeof(IFuid));

                        if (!data->namelist)
                            return (E_NOMEM);

                        (*(SPfrontEnd->IFnewUid))(ckt,
                            &(data->namelist[data->numPlots++]),
                            (IFuid)NULL,name,UID_OTHER,(GENERIC **)NULL);
                        /* we've added one more plot */
                    }
                }
            }
        }
        return (OK);
    }

    if (operation == N_CALC) {

        if (mode == N_DENS) {

            for ( ; model; model = model->DIOnextModel) {
                for (inst = model->DIOinstances; inst;
                        inst = inst->DIOnextInstance) {

                    NevalSrc(&noizDens[DIORSNOIZ],&lnNdens[DIORSNOIZ],
                        ckt,THERMNOISE,inst->DIOposPrimeNode,
                        inst->DIOposNode,
                        model->DIOconductance * inst->DIOarea);

                    NevalSrc(&noizDens[DIOIDNOIZ],&lnNdens[DIOIDNOIZ],
                        ckt,SHOTNOISE,inst->DIOposPrimeNode,
                        inst->DIOnegNode,inst->DIOcd);

                    NevalSrc(&noizDens[DIOFLNOIZ],(double*)NULL,
                        ckt,N_GAIN,inst->DIOposPrimeNode,
                        inst->DIOnegNode,(double)0.0);

                    noizDens[DIOFLNOIZ] *= model->DIOfNcoef * 
                        exp(model->DIOfNexp * log(MAX(FABS(
                        inst->DIOcd),N_MINLOG))) / data->freq;

                    lnNdens[DIOFLNOIZ] = 
                        log(MAX(noizDens[DIOFLNOIZ],N_MINLOG));

                    noizDens[DIOTOTNOIZ] = noizDens[DIORSNOIZ] +
                        noizDens[DIOIDNOIZ] + noizDens[DIOFLNOIZ];

                    lnNdens[DIOTOTNOIZ] =
                        log(MAX(noizDens[DIOTOTNOIZ], N_MINLOG));

                    *OnDens += noizDens[DIOTOTNOIZ];

                    if (data->delFreq == 0.0) { 

                        /* if we haven't done any previous integration,
                         * we need to initialize our "history" variables
                         */

                        for (i = 0; i < DIONSRCS; i++) {
                            inst->DIOnVar[LNLSTDENS][i] = lnNdens[i];
                        }

                        /* clear out our integration variables if it's the
                         * first pass
                         */

                        if (data->freq ==
                                ((NOISEAN*)ckt->CKTcurJob)->AC.fstart) {

                            for (i = 0; i < DIONSRCS; i++) {
                                inst->DIOnVar[OUTNOIZ][i] = 0.0;
                                inst->DIOnVar[INNOIZ][i] = 0.0;
                            }
                        }
                    }
                    else {

                        /* data->delFreq != 0.0 (we have to integrate) */

                        /* To insure accurracy, we have to integrate each
                         * component separately
                         */

                        for (i = 0; i < DIONSRCS; i++) {
                            if (i != DIOTOTNOIZ) {
                                tempOnoise =
                                    Nintegrate(noizDens[i],lnNdens[i],
                                    inst->DIOnVar[LNLSTDENS][i],data);
                                tempInoise =
                                    Nintegrate(noizDens[i] * data->GainSqInv,
                                    lnNdens[i] + data->lnGainInv,
                                    inst->DIOnVar[LNLSTDENS][i] +
                                    data->lnGainInv,data);
                                inst->DIOnVar[LNLSTDENS][i] = lnNdens[i];
                                data->outNoiz += tempOnoise;
                                data->inNoise += tempInoise;
                                if (((NOISEAN*)ckt->CKTcurJob)->NStpsSm
                                        != 0) {
                                    inst->DIOnVar[OUTNOIZ][i] += tempOnoise;
                                    inst->DIOnVar[OUTNOIZ][DIOTOTNOIZ] +=
                                        tempOnoise;
                                    inst->DIOnVar[INNOIZ][i] += tempInoise;
                                    inst->DIOnVar[INNOIZ][DIOTOTNOIZ] +=
                                        tempInoise;
                                }
                            }
                        }
                    }
                    if (data->prtSummary) {
                        for (i = 0; i < DIONSRCS; i++) {
                            /* print a summary report */
                            data->outpVector[data->outNumber++] =
                                noizDens[i];
                        }
                    }
                }
            }
            return (OK);
        }

        if (mode == INT_NOIZ) {

            /* already calculated, just output */

            if (((NOISEAN*)ckt->CKTcurJob)->NStpsSm == 0)
                return (OK);

            for ( ; model; model = model->DIOnextModel) {
                for (inst = model->DIOinstances; inst;
                        inst = inst->DIOnextInstance) {

                    for (i = 0; i < DIONSRCS; i++) {
                        data->outpVector[data->outNumber++] =
                            inst->DIOnVar[OUTNOIZ][i];
                        data->outpVector[data->outNumber++] =
                            inst->DIOnVar[INNOIZ][i];
                    }
                }
            }
        }
    }

    return (OK);
}
