/***************************************************************************
JSPICE3 adaptation of Spice3f2 - Copyright (c) Stephen R. Whiteley 1992
Copyright 1990 Regents of the University of California.  All rights reserved.
Authors: 1987 Gary W. Ng
         1993 Stephen R. Whiteley
****************************************************************************/

#include "spice.h"
#include <stdio.h>
#include <math.h>
#include "mesdefs.h"
#include "noisedef.h"
#include "fteconst.h"
#include "iferrmsg.h"
#include "util.h"

/*
 * MESnoise (mode, operation, firstModel, ckt, data, OnDens)
 *    This routine names and evaluates all of the noise sources
 *    associated with MESFET's.  It starts with the model *firstModel and
 *    traverses all of its insts.  It then proceeds to any other models
 *    on the linked list.  The total output noise density generated by
 *    all of the MESFET's is summed with the variable "OnDens".
 */

/* define the names of the noise sources */
static char *MESnNames[MESNSRCS] = {
    /* Note that we have to keep the order
     * consistent with the index definitions
     * in MESdefs.h
     */
    ".rd",      /* noise due to rd */
    ".rs",      /* noise due to rs */
    ".id",      /* noise due to id */
    ".1overf",  /* flicker (1/f) noise */
    ""          /* total transistor noise */
};


int
MESnoise (mode, operation, genmodel, ckt, dataptr, OnDens)

int mode;
int operation;
GENmodel *genmodel;
CKTcircuit *ckt;
GENERIC *dataptr;
double *OnDens;
{
    Ndata *data = (Ndata*)dataptr;
    MESmodel *model = (MESmodel *) genmodel;
    MESinstance *inst;
    char name[N_MXVLNTH];
    double tempOnoise;
    double tempInoise;
    double noizDens[MESNSRCS];
    double lnNdens[MESNSRCS];
    int error;
    int i;


    if (operation == N_OPEN) {

        /* see if we have to to produce a summary report
         * if so, name all the noise generators
         */
        if (((NOISEAN*)ckt->CKTcurJob)->NStpsSm == 0)
            return (OK);

        if (mode == N_DENS) {

            for ( ; model; model = model->MESnextModel) {
                for (inst = model->MESinstances; inst;
                        inst = inst->MESnextInstance) {

                    for (i = 0; i < MESNSRCS; i++) {
                        (void)sprintf(name,
                            "onoise.%s%s",inst->MESname,MESnNames[i]);

                        data->namelist = (IFuid *)
                            trealloc((char *)data->namelist,
                                (data->numPlots + 1)*sizeof(IFuid));

                        if (!data->namelist)
                            return (E_NOMEM);

                        (*(SPfrontEnd->IFnewUid))(ckt,
                            &(data->namelist[data->numPlots++]),
                            (IFuid)NULL,name,UID_OTHER,(GENERIC **)NULL);
                        /* we've added one more plot */
                    }
                }
            }
            return (OK);
        }

        if (mode == INT_NOIZ) {

            for ( ; model; model = model->MESnextModel) {
                for (inst = model->MESinstances; inst;
                        inst = inst->MESnextInstance) {

                    for (i = 0; i < MESNSRCS; i++) {
                        (void)sprintf(name,
                            "onoise_total.%s%s",inst->MESname,MESnNames[i]);

                        data->namelist = (IFuid *)
                            trealloc((char *)data->namelist,
                                (data->numPlots + 1)*sizeof(IFuid));

                        if (!data->namelist)
                            return (E_NOMEM);

                        (*(SPfrontEnd->IFnewUid))(ckt,
                            &(data->namelist[data->numPlots++]),
                            (IFuid)NULL,name,UID_OTHER,(GENERIC **)NULL);
                        /* we've added one more plot */

                        (void)sprintf(name,
                            "inoise_total.%s%s",inst->MESname,MESnNames[i]);

                        data->namelist = (IFuid *)
                            trealloc((char *)data->namelist,
                                (data->numPlots + 1)*sizeof(IFuid));

                        if (!data->namelist)
                            return (E_NOMEM);

                        (*(SPfrontEnd->IFnewUid))(ckt,
                            &(data->namelist[data->numPlots++]),
                            (IFuid)NULL,name,UID_OTHER,(GENERIC **)NULL);
                        /* we've added one more plot */
                    }
                }
            }
        }
        return (OK);
    }

    if (operation == N_CALC) {

        if (mode == N_DENS) {

            for ( ; model; model = model->MESnextModel) {
                for (inst = model->MESinstances; inst;
                        inst = inst->MESnextInstance) {

                    NevalSrc(&noizDens[MESRDNOIZ],&lnNdens[MESRDNOIZ],
                        ckt,THERMNOISE,inst->MESdrainPrimeNode,
                        inst->MESdrainNode,
                        model->MESdrainConduct * inst->MESarea);

                    NevalSrc(&noizDens[MESRSNOIZ],&lnNdens[MESRSNOIZ],
                        ckt,THERMNOISE,inst->MESsourcePrimeNode,
                        inst->MESsourceNode,
                        model->MESsourceConduct * inst->MESarea);

                    NevalSrc(&noizDens[MESIDNOIZ],&lnNdens[MESIDNOIZ],
                        ckt,THERMNOISE,inst->MESdrainPrimeNode,
                        inst->MESsourcePrimeNode,
                        (2.0/3.0 * FABS(*(ckt->CKTstate0 + inst->MESgm))));

                    NevalSrc(&noizDens[MESFLNOIZ],(double*)NULL,ckt,
                        N_GAIN,inst->MESdrainPrimeNode,
                        inst->MESsourcePrimeNode,(double)0.0);

                    noizDens[MESFLNOIZ] *= model->MESfNcoef * 
                        exp(model->MESfNexp *
                        log(MAX(FABS(*(ckt->CKTstate0 +
                        inst->MEScd)),N_MINLOG))) / data->freq;

                    lnNdens[MESFLNOIZ] = 
                        log(MAX(noizDens[MESFLNOIZ],N_MINLOG));

                    noizDens[MESTOTNOIZ] = noizDens[MESRDNOIZ] +
                        noizDens[MESRSNOIZ] + noizDens[MESIDNOIZ] +
                        noizDens[MESFLNOIZ];

                    lnNdens[MESTOTNOIZ] = 
                        log(MAX(noizDens[MESTOTNOIZ], N_MINLOG));

                    *OnDens += noizDens[MESTOTNOIZ];

                    if (data->delFreq == 0.0) { 

                        /* if we haven't done any previous integration,
                         * we need to initialize our "history" variables
                         */

                        for (i = 0; i < MESNSRCS; i++) {
                            inst->MESnVar[LNLSTDENS][i] = lnNdens[i];
                        }

                        /* clear out our integration variables if it's
                         * the first pass
                         */

                        if (data->freq ==
                                ((NOISEAN*)ckt->CKTcurJob)->AC.fstart) {

                            for (i = 0; i < MESNSRCS; i++) {
                                inst->MESnVar[OUTNOIZ][i] = 0.0;
                                inst->MESnVar[INNOIZ][i] = 0.0;
                            }
                        }
                    }
                    else {
                        /* data->delFreq != 0.0 (we have to integrate) */

                        for (i = 0; i < MESNSRCS; i++) {
                            if (i != MESTOTNOIZ) {

                                tempOnoise =
                                    Nintegrate(noizDens[i], lnNdens[i],
                                        inst->MESnVar[LNLSTDENS][i], data);
                                tempInoise =
                                    Nintegrate(noizDens[i] * data->GainSqInv,
                                    lnNdens[i] + data->lnGainInv,
                                    inst->MESnVar[LNLSTDENS][i] +
                                    data->lnGainInv,data);
                                inst->MESnVar[LNLSTDENS][i] = lnNdens[i];
                                data->outNoiz += tempOnoise;
                                data->inNoise += tempInoise;
                                if (((NOISEAN*)ckt->CKTcurJob)->NStpsSm
                                        != 0) {
                                    inst->MESnVar[OUTNOIZ][i] += tempOnoise;
                                    inst->MESnVar[OUTNOIZ][MESTOTNOIZ] +=
                                        tempOnoise;
                                    inst->MESnVar[INNOIZ][i] += tempInoise;
                                    inst->MESnVar[INNOIZ][MESTOTNOIZ] +=
                                        tempInoise;
                                }
                            }
                        }
                    }
                    if (data->prtSummary) {
                        /* print a summary report */
                        for (i = 0; i < MESNSRCS; i++) {
                            data->outpVector[data->outNumber++] =
                                noizDens[i];
                        }
                    }
                }
            }
            return (OK);
        }

        if (mode == INT_NOIZ) {
            /* already calculated, just output */

            if (((NOISEAN*)ckt->CKTcurJob)->NStpsSm == 0)
                return (OK);

            for ( ; model; model = model->MESnextModel) {
                for (inst = model->MESinstances; inst;
                        inst = inst->MESnextInstance) {

                    for (i = 0; i < MESNSRCS; i++) {
                        data->outpVector[data->outNumber++] =
                            inst->MESnVar[OUTNOIZ][i];
                        data->outpVector[data->outNumber++] =
                            inst->MESnVar[INNOIZ][i];
                    }
                }
            }
        }
    }

    return (OK);
}
            

